@@ -1,14 +1,50 @@
 /* MiniDLNA project
- * http://minidlna.sourceforge.net/
- * (c) 2008-2009 Justin Maggard
  *
- * This software is subject to the conditions detailed
- * in the LICENCE file provided within the distribution 
+ * http://sourceforge.net/projects/minidlna/
  *
- * Portions of the code from the MiniUPnP Project
- * (c) Thomas Bernard licensed under BSD revised license
- * detailed in the LICENSE.miniupnpd file provided within
- * the distribution.
+ * MiniDLNA media server
+ * Copyright (C) 2008-2009  Justin Maggard
+ *
+ * This file is part of MiniDLNA.
+ *
+ * MiniDLNA is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License version 2 as
+ * published by the Free Software Foundation.
+ *
+ * MiniDLNA is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with MiniDLNA. If not, see <http://www.gnu.org/licenses/>.
+ *
+ * Portions of the code from the MiniUPnP project:
+ *
+ * Copyright (c) 2006-2007, Thomas Bernard
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
+ *     * Redistributions of source code must retain the above copyright
+ *       notice, this list of conditions and the following disclaimer.
+ *     * Redistributions in binary form must reproduce the above copyright
+ *       notice, this list of conditions and the following disclaimer in the
+ *       documentation and/or other materials provided with the distribution.
+ *     * The name of the author may not be used to endorse or promote products
+ *       derived from this software without specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+ * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
+ * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
+ * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
+ * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
+ * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
+ * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
+ * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <stdlib.h>
 #include <unistd.h>
@@ -45,6 +81,7 @@
 #endif
 //#define MAX_BUFFER_SIZE 4194304 // 4MB -- Too much?
 #define MAX_BUFFER_SIZE 2147483647 // 2GB -- Too much?
+#define MIN_BUFFER_SIZE 65536
 
 #include "icons.c"
 
@@ -80,16 +117,14 @@ Delete_upnphttp(struct upnphttp * h)
 	{
 		if(h->socket >= 0)
 			CloseSocket_upnphttp(h);
-		if(h->req_buf)
-			free(h->req_buf);
-		if(h->res_buf)
-			free(h->res_buf);
+		free(h->req_buf);
+		free(h->res_buf);
 		free(h);
 	}
 }
 
 int
-SearchClientCache(struct in_addr addr)
+SearchClientCache(struct in_addr addr, int quiet)
 {
 	int i;
 	for( i=0; i<CLIENT_CACHE_SLOTS; i++ )
@@ -113,7 +148,9 @@ SearchClientCache(struct in_addr addr)
 					return -1;
 				}
 			}
-			DPRINTF(E_DEBUG, L_HTTP, "Client found in cache. [type %d/entry %d]\n", clients[i].type, i);
+			if( !quiet )
+				DPRINTF(E_DEBUG, L_HTTP, "Client found in cache. [type %d/entry %d]\n",
+					clients[i].type, i);
 			return i;
 		}
 	}
@@ -218,15 +255,32 @@ intervening space) by either an integer or the keyword "infinite". */
 					p++;
 				if(strncasecmp(p, "bytes=", 6)==0) {
 					h->reqflags |= FLAG_RANGE;
-					h->req_RangeEnd = atoll(index(p+6, '-')+1);
-					h->req_RangeStart = atoll(p+6);
+					h->req_RangeStart = strtoll(p+6, &colon, 10);
+					h->req_RangeEnd = colon ? atoll(colon+1) : 0;
 					DPRINTF(E_DEBUG, L_HTTP, "Range Start-End: %lld - %lld\n",
 					       h->req_RangeStart, h->req_RangeEnd?h->req_RangeEnd:-1);
 				}
 			}
 			else if(strncasecmp(line, "Host", 4)==0)
 			{
+				int i;
 				h->reqflags |= FLAG_HOST;
+				p = colon + 1;
+				while(isspace(*p))
+					p++;
+				for(n = 0; n<n_lan_addr; n++)
+				{
+					for(i=0; lan_addr[n].str[i]; i++)
+					{
+						if(lan_addr[n].str[i] != p[i])
+							break;
+					}
+					if(!lan_addr[n].str[i])
+					{
+						h->iface = n;
+						break;
+					}
+				}
 			}
 			else if(strncasecmp(line, "User-Agent", 10)==0)
 			{
@@ -240,6 +294,7 @@ intervening space) by either an integer or the keyword "infinite". */
 				{
 					h->req_client = EXbox;
 					h->reqflags |= FLAG_MIME_AVI_AVI;
+					h->reqflags |= FLAG_MS_PFS;
 				}
 				else if(strncmp(p, "PLAYSTATION", 11)==0)
 				{
@@ -247,13 +302,19 @@ intervening space) by either an integer or the keyword "infinite". */
 					h->reqflags |= FLAG_DLNA;
 					h->reqflags |= FLAG_MIME_AVI_DIVX;
 				}
-				else if(strncmp(p, "SamsungWiselinkPro", 18)==0 ||
-				        strncmp(p, "SEC_HHP_", 8)==0)
+				else if(strncmp(p, "SEC_HHP_", 8)==0)
 				{
 					h->req_client = ESamsungTV;
+					h->reqflags |= FLAG_SAMSUNG;
+					h->reqflags |= FLAG_DLNA;
+					h->reqflags |= FLAG_NO_RESIZE;
+				}
+				else if(strncmp(p, "SamsungWiselinkPro", 18)==0)
+				{
+					h->req_client = ESamsungSeriesA;
+					h->reqflags |= FLAG_SAMSUNG;
 					h->reqflags |= FLAG_DLNA;
 					h->reqflags |= FLAG_NO_RESIZE;
-					//h->reqflags |= FLAG_MIME_AVI_DIVX;
 				}
 				else if(strstrc(p, "bridgeCo-DMP/3", '\r'))
 				{
@@ -269,6 +330,26 @@ intervening space) by either an integer or the keyword "infinite". */
 					h->req_client = EPopcornHour;
 					h->reqflags |= FLAG_MIME_FLAC_FLAC;
 				}
+				else if(strstrc(p, "Microsoft-IPTV-Client", '\r'))
+				{
+					h->req_client = EMediaRoom;
+					h->reqflags |= FLAG_MS_PFS;
+				}
+				else if(strstrc(p, "LGE_DLNA_SDK", '\r'))
+				{
+					h->req_client = ELGDevice;
+					h->reqflags |= FLAG_DLNA;
+				}
+				else if(strncmp(p, "Verismo,", 8)==0)
+				{
+					h->req_client = ENetgearEVA2000;
+					h->reqflags |= FLAG_MS_PFS;
+				}
+				else if(strstrc(p, "UPnP/1.0 DLNADOC/1.50 Intel_SDK_for_UPnP_devices/1.2", '\r'))
+				{
+					h->req_client = EToshibaTV;
+					h->reqflags |= FLAG_DLNA;
+				}
 				else if(strstrc(p, "DLNADOC/1.50", '\r'))
 				{
 					h->req_client = EStandardDLNA150;
@@ -279,17 +360,36 @@ intervening space) by either an integer or the keyword "infinite". */
 			else if(strncasecmp(line, "X-AV-Client-Info", 16)==0)
 			{
 				/* Skip client detection if we already detected it. */
-				if( h->req_client )
+				if( h->req_client && h->req_client < EStandardDLNA150 )
 					goto next_header;
 				p = colon + 1;
 				while(isspace(*p))
 					p++;
-				if(strstr(p, "PLAYSTATION 3"))
+				if(strstrc(p, "PLAYSTATION 3", '\r'))
 				{
 					h->req_client = EPS3;
 					h->reqflags |= FLAG_DLNA;
 					h->reqflags |= FLAG_MIME_AVI_DIVX;
 				}
+				/* X-AV-Client-Info: av=5.0; cn="Sony Corporation"; mn="Blu-ray Disc Player"; mv="2.0" */
+				/* X-AV-Client-Info: av=5.0; cn="Sony Corporation"; mn="BLU-RAY HOME THEATRE SYSTEM"; mv="2.0"; */
+				/* Sony SMP-100 needs the same treatment as their BDP-S370 */
+				/* X-AV-Client-Info: av=5.0; cn="Sony Corporation"; mn="Media Player"; mv="2.0" */
+				else if(strstrc(p, "Blu-ray Disc Player", '\r') ||
+				        strstrc(p, "BLU-RAY HOME THEATRE SYSTEM", '\r') ||
+				        strstrc(p, "Media Player", '\r'))
+				{
+					h->req_client = ESonyBDP;
+					h->reqflags |= FLAG_DLNA;
+				}
+				/* X-AV-Client-Info: av=5.0; cn="Sony Corporation"; mn="BRAVIA KDL-40EX503"; mv="1.7"; */
+				/* X-AV-Client-Info: av=5.0; hn=""; cn="Sony Corporation"; mn="INTERNET TV NSX-40GT 1"; mv="0.1"; */
+				else if(strstrc(p, "BRAVIA", '\r') ||
+				        strstrc(p, "INTERNET TV", '\r'))
+				{
+					h->req_client = ESonyBravia;
+					h->reqflags |= FLAG_DLNA;
+				}
 			}
 			else if(strncasecmp(line, "Transfer-Encoding", 17)==0)
 			{
@@ -375,7 +475,7 @@ next_header:
 	/* If the client type wasn't found, search the cache.
 	 * This is done because a lot of clients like to send a
 	 * different User-Agent with different types of requests. */
-	n = SearchClientCache(h->clientaddr);
+	n = SearchClientCache(h->clientaddr, 0);
 	if( h->req_client )
 	{
 		/* Add this client to the cache if it's not there already. */
@@ -394,7 +494,7 @@ next_header:
 				break;
 			}
 		}
-		else if( (n < EStandardDLNA150) && (h->req_client == EStandardDLNA150) )
+		else if( (clients[n].type < EStandardDLNA150) && (h->req_client == EStandardDLNA150) )
 		{
 			/* If we know the client and our new detection is generic, use our cached info */
 			h->reqflags |= clients[n].flags;
@@ -473,7 +573,7 @@ Send416(struct upnphttp * h)
 }
 
 /* very minimalistic 500 error message */
-static void
+void
 Send500(struct upnphttp * h)
 {
 	static const char body500[] = 
@@ -505,7 +605,7 @@ Send501(struct upnphttp * h)
 static const char *
 findendheaders(const char * s, int len)
 {
-	while(len-->0)
+	while(len-- > 0)
 	{
 		if(s[0]=='\r' && s[1]=='\n' && s[2]=='\r' && s[3]=='\n')
 			return s;
@@ -525,13 +625,9 @@ sendXMLdesc(struct upnphttp * h, char * (f)(int *))
 	{
 		DPRINTF(E_ERROR, L_HTTP, "Failed to generate XML description\n");
 		Send500(h);
-		free(desc);
 		return;
 	}
-	else
-	{
-		BuildResp_upnphttp(h, desc, len);
-	}
+	BuildResp_upnphttp(h, desc, len);
 	SendResp_upnphttp(h);
 	CloseSocket_upnphttp(h);
 	free(desc);
@@ -556,7 +652,7 @@ ProcessHTTPPOST_upnphttp(struct upnphttp * h)
 		{
 			static const char err400str[] =
 				"<html><body>Bad request</body></html>";
-			DPRINTF(E_WARN, L_HTTP, "No SOAPAction in HTTP headers");
+			DPRINTF(E_WARN, L_HTTP, "No SOAPAction in HTTP headers\n");
 			h->respflags = FLAG_HTML;
 			BuildResp2_upnphttp(h, 400, "Bad Request",
 			                    err400str, sizeof(err400str) - 1);
@@ -741,9 +837,10 @@ ProcessHttpQuery_upnphttp(struct upnphttp * h)
 			/* If it's a Xbox360, we might need a special friendly_name to be recognized */
 			if( (h->req_client == EXbox) && !strchr(friendly_name, ':') )
 			{
-				strncat(friendly_name, ": 1", FRIENDLYNAME_MAX_LEN-4);
+				i = strlen(friendly_name);
+				snprintf(friendly_name+i, FRIENDLYNAME_MAX_LEN-i, ": 1");
 				sendXMLdesc(h, genRootDesc);
-				friendly_name[strlen(friendly_name)-3] = '\0';
+				friendly_name[i] = '\0';
 			}
 			else
 			{
@@ -787,13 +884,14 @@ ProcessHttpQuery_upnphttp(struct upnphttp * h)
 				}
 				else
 				{
-					printf("Invalid TiVo request! %s\n", HttpUrl+12);
+					DPRINTF(E_WARN, L_HTTP, "Invalid TiVo request! %s\n", HttpUrl+12);
 					Send404(h);
 				}
 			}
 			else
 			{
-				printf("TiVo request with out TiVo support enabled! %s\n", HttpUrl+12);
+				DPRINTF(E_WARN, L_HTTP, "TiVo request with out TiVo support enabled! %s\n",
+					HttpUrl+12);
 				Send404(h);
 			}
 		}
@@ -1027,7 +1125,7 @@ SendResp_upnphttp(struct upnphttp * h)
 	n = send(h->socket, h->res_buf, h->res_buflen, 0);
 	if(n<0)
 	{
-		DPRINTF(E_ERROR, L_HTTP, "send(res_buf): %s", strerror(errno));
+		DPRINTF(E_ERROR, L_HTTP, "send(res_buf): %s\n", strerror(errno));
 	}
 	else if(n < h->res_buflen)
 	{
@@ -1065,32 +1163,60 @@ send_file(struct upnphttp * h, int sendfd, off_t offset, off_t end_offset)
 {
 	off_t send_size;
 	off_t ret;
+	char *buf = NULL;
+	int try_sendfile = 1;
 
 	while( offset < end_offset )
 	{
-		send_size = ( ((end_offset - offset) < MAX_BUFFER_SIZE) ? (end_offset - offset + 1) : MAX_BUFFER_SIZE);
-		ret = sendfile(h->socket, sendfd, &offset, send_size);
-		if( ret == -1 )
+		if( try_sendfile )
 		{
-			DPRINTF(E_DEBUG, L_HTTP, "sendfile error :: error no. %d [%s]\n", errno, strerror(errno));
+			send_size = ( ((end_offset - offset) < MAX_BUFFER_SIZE) ? (end_offset - offset + 1) : MAX_BUFFER_SIZE);
+			ret = sendfile(h->socket, sendfd, &offset, send_size);
+			if( ret == -1 )
+			{
+				DPRINTF(E_DEBUG, L_HTTP, "sendfile error :: error no. %d [%s]\n", errno, strerror(errno));
+				/* If sendfile isn't supported on the filesystem, don't bother trying to use it again. */
+				if( errno == EOVERFLOW || errno == EINVAL )
+					try_sendfile = 0;
+				else if( errno != EAGAIN )
+					break;
+			}
+			else
+			{
+				//DPRINTF(E_DEBUG, L_HTTP, "sent %lld bytes to %d. offset is now %lld.\n", ret, h->socket, offset);
+				continue;
+			}
+		}
+		/* Fall back to regular I/O */
+		if( !buf )
+			buf = malloc(MIN_BUFFER_SIZE);
+		send_size = ( ((end_offset - offset) < MIN_BUFFER_SIZE) ? (end_offset - offset + 1) : MIN_BUFFER_SIZE);
+		lseek(sendfd, offset, SEEK_SET);
+		ret = read(sendfd, buf, send_size);
+		if( ret == -1 ) {
+			DPRINTF(E_DEBUG, L_HTTP, "read error :: error no. %d [%s]\n", errno, strerror(errno));
 			if( errno != EAGAIN )
 				break;
 		}
-		/*else
-		{
-			DPRINTF(E_DEBUG, L_HTTP, "sent %lld bytes to %d. offset is now %lld.\n", ret, h->socket, offset);
-		}*/
+		ret = write(h->socket, buf, ret);
+		if( ret == -1 ) {
+			DPRINTF(E_DEBUG, L_HTTP, "write error :: error no. %d [%s]\n", errno, strerror(errno));
+			if( errno != EAGAIN )
+				break;
+		}
+		offset+=ret;
 	}
+	free(buf);
 }
 
 void
 SendResp_icon(struct upnphttp * h, char * icon)
 {
-	char * header;
-	char * data;
-	int size, ret;
-	char mime[12];
+	char header[512];
+	char mime[12] = "image/";
 	char date[30];
+	char *data;
+	int size, ret;
 	time_t curtime = time(NULL);
 
 	if( strcmp(icon, "sm.png") == 0 )
@@ -1098,28 +1224,28 @@ SendResp_icon(struct upnphttp * h, char * icon)
 		DPRINTF(E_DEBUG, L_HTTP, "Sending small PNG icon\n");
 		data = (char *)png_sm;
 		size = sizeof(png_sm)-1;
-		strcpy(mime, "image/png");
+		strcpy(mime+6, "png");
 	}
 	else if( strcmp(icon, "lrg.png") == 0 )
 	{
 		DPRINTF(E_DEBUG, L_HTTP, "Sending large PNG icon\n");
 		data = (char *)png_lrg;
 		size = sizeof(png_lrg)-1;
-		strcpy(mime, "image/png");
+		strcpy(mime+6, "png");
 	}
 	else if( strcmp(icon, "sm.jpg") == 0 )
 	{
 		DPRINTF(E_DEBUG, L_HTTP, "Sending small JPEG icon\n");
 		data = (char *)jpeg_sm;
 		size = sizeof(jpeg_sm)-1;
-		strcpy(mime, "image/jpeg");
+		strcpy(mime+6, "jpeg");
 	}
 	else if( strcmp(icon, "lrg.jpg") == 0 )
 	{
 		DPRINTF(E_DEBUG, L_HTTP, "Sending large JPEG icon\n");
 		data = (char *)jpeg_lrg;
 		size = sizeof(jpeg_lrg)-1;
-		strcpy(mime, "image/jpeg");
+		strcpy(mime+6, "jpeg");
 	}
 	else
 	{
@@ -1128,38 +1254,33 @@ SendResp_icon(struct upnphttp * h, char * icon)
 		return;
 	}
 
-
 	strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
-	ret = asprintf(&header, "HTTP/1.1 200 OK\r\n"
-	                        "Content-Type: %s\r\n"
-	                        "Content-Length: %d\r\n"
-	                        "Connection: close\r\n"
-	                        "Date: %s\r\n"
-	                        "Server: " MINIDLNA_SERVER_STRING "\r\n\r\n",
-	                        mime, size, date);
+	ret = snprintf(header, sizeof(header), "HTTP/1.1 200 OK\r\n"
+	                                       "Content-Type: %s\r\n"
+	                                       "Content-Length: %d\r\n"
+	                                       "Connection: close\r\n"
+	                                       "Date: %s\r\n"
+	                                       "Server: " MINIDLNA_SERVER_STRING "\r\n\r\n",
+	                                       mime, size, date);
 
-	if( (send_data(h, header, ret, MSG_MORE) == 0) && (h->req_command != EHead) )
+	if( send_data(h, header, ret, MSG_MORE) == 0 )
 	{
-		send_data(h, data, size, 0);
+ 		if( h->req_command != EHead )
+			send_data(h, data, size, 0);
 	}
-	free(header);
 }
 
 void
 SendResp_albumArt(struct upnphttp * h, char * object)
 {
-	char header[1500];
-	char sql_buf[256];
-	char **result;
-	int rows = 0;
+	char header[512];
 	char *path;
 	char *dash;
 	char date[30];
 	time_t curtime = time(NULL);
-	off_t offset = 0, size;
-	int sendfh;
-
-	memset(header, 0, 1500);
+	off_t size;
+	int fd;
+	int ret;
 
 	if( h->reqflags & FLAG_XFERSTREAMING || h->reqflags & FLAG_RANGE )
 	{
@@ -1171,98 +1292,81 @@ SendResp_albumArt(struct upnphttp * h, char * object)
 	dash = strchr(object, '-');
 	if( dash )
 		*dash = '\0';
-	sprintf(sql_buf, "SELECT PATH from ALBUM_ART where ID = %s", object);
-	sql_get_table(db, sql_buf, &result, &rows, NULL);
-	if( !rows )
+
+	path = sql_get_text_field(db, "SELECT PATH from ALBUM_ART where ID = '%s'", object);
+	if( !path )
 	{
 		DPRINTF(E_WARN, L_HTTP, "ALBUM_ART ID %s not found, responding ERROR 404\n", object);
 		Send404(h);
-		goto error;
+		return;
 	}
-	path = result[1];
 	DPRINTF(E_INFO, L_HTTP, "Serving album art ID: %s [%s]\n", object, path);
 
-	if( access(path, F_OK) == 0 )
-	{
-		strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
+	strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
 
-		sendfh = open(path, O_RDONLY);
-		if( sendfh < 0 ) {
-			DPRINTF(E_ERROR, L_HTTP, "Error opening %s\n", path);
-			goto error;
-		}
-		size = lseek(sendfh, 0, SEEK_END);
-		lseek(sendfh, 0, SEEK_SET);
-
-		sprintf(header, "HTTP/1.1 200 OK\r\n"
-				"Content-Type: image/jpeg\r\n"
-				"Content-Length: %jd\r\n"
-				"Connection: close\r\n"
-				"Date: %s\r\n"
-				"EXT:\r\n"
-				"realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
-				"contentFeatures.dlna.org: DLNA.ORG_PN=JPEG_TN\r\n"
-				"Server: " MINIDLNA_SERVER_STRING "\r\n",
-				size, date);
-
-		if( h->reqflags & FLAG_XFERBACKGROUND )
-		{
-			strcat(header, "transferMode.dlna.org: Background\r\n\r\n");
-		}
-		else //if( h->reqflags & FLAG_XFERINTERACTIVE )
-		{
-			strcat(header, "transferMode.dlna.org: Interactive\r\n\r\n");
-		}
-
-
-		if( (send_data(h, header, strlen(header), MSG_MORE) == 0) && (h->req_command != EHead) && (sendfh > 0) )
-		{
-			send_file(h, sendfh, offset, size);
-		}
-		close(sendfh);
+	fd = open(path, O_RDONLY);
+	if( fd < 0 ) {
+		DPRINTF(E_ERROR, L_HTTP, "Error opening %s\n", path);
+		sqlite3_free(path);
+		Send404(h);
+		return;
 	}
-	error:
-	sqlite3_free_table(result);
+	sqlite3_free(path);
+	size = lseek(fd, 0, SEEK_END);
+	lseek(fd, 0, SEEK_SET);
+
+	ret = snprintf(header, sizeof(header), "HTTP/1.1 200 OK\r\n"
+	                                       "Content-Type: image/jpeg\r\n"
+	                                       "Content-Length: %jd\r\n"
+	                                       "Connection: close\r\n"
+	                                       "Date: %s\r\n"
+	                                       "EXT:\r\n"
+	                                       "realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
+	                                       "contentFeatures.dlna.org: DLNA.ORG_PN=JPEG_TN\r\n"
+	                                       "Server: " MINIDLNA_SERVER_STRING "\r\n"
+	                                       "transferMode.dlna.org: %s\r\n\r\n",
+	                                       (intmax_t)size, date,
+	                                       (h->reqflags & FLAG_XFERBACKGROUND) ? "Background" : "Interactive");
+
+	if( send_data(h, header, ret, MSG_MORE) == 0 )
+	{
+ 		if( h->req_command != EHead )
+			send_file(h, fd, 0, size-1);
+	}
+	close(fd);
 }
 
 void
 SendResp_caption(struct upnphttp * h, char * object)
 {
-	char header[1500];
-	char sql_buf[256];
-	char **result;
-	int rows = 0;
+	char header[512];
 	char *path;
 	char date[30];
 	time_t curtime = time(NULL);
-	off_t offset = 0, size;
-	int sendfh, ret;
-
-	memset(header, 0, 1500);
+	off_t size;
+	int fd, ret;
 
 	strip_ext(object);
-	sprintf(sql_buf, "SELECT PATH from CAPTIONS where ID = %s", object);
-	sql_get_table(db, sql_buf, &result, &rows, NULL);
-	if( !rows )
+	path = sql_get_text_field(db, "SELECT PATH from CAPTIONS where ID = %s", object);
+	if( !path )
 	{
 		DPRINTF(E_WARN, L_HTTP, "CAPTION ID %s not found, responding ERROR 404\n", object);
 		Send404(h);
-		goto error;
+		return;
 	}
-	path = result[1];
 	DPRINTF(E_INFO, L_HTTP, "Serving caption ID: %s [%s]\n", object, path);
 
-	if( access(path, F_OK) != 0 )
-		goto error;
-
-	strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
-	sendfh = open(path, O_RDONLY);
-	if( sendfh < 0 ) {
+	fd = open(path, O_RDONLY);
+	if( fd < 0 ) {
 		DPRINTF(E_ERROR, L_HTTP, "Error opening %s\n", path);
-		goto error;
+		sqlite3_free(path);
+		Send404(h);
+		return;
 	}
-	size = lseek(sendfh, 0, SEEK_END);
-	lseek(sendfh, 0, SEEK_SET);
+	sqlite3_free(path);
+	size = lseek(fd, 0, SEEK_END);
+	lseek(fd, 0, SEEK_SET);
+	strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
 
 	ret = snprintf(header, sizeof(header), "HTTP/1.1 200 OK\r\n"
 	                                       "Content-Type: smi/caption\r\n"
@@ -1271,33 +1375,27 @@ SendResp_caption(struct upnphttp * h, char * object)
 	                                       "Date: %s\r\n"
 	                                       "EXT:\r\n"
 	                                       "Server: " MINIDLNA_SERVER_STRING "\r\n\r\n",
-	                                       size, date);
+	                                       (intmax_t)size, date);
 
-	if( (send_data(h, header, ret, MSG_MORE) == 0) && (h->req_command != EHead) && (sendfh > 0) )
+	if( send_data(h, header, ret, MSG_MORE) == 0 )
 	{
-		send_file(h, sendfh, offset, size);
+ 		if( h->req_command != EHead )
+			send_file(h, fd, 0, size-1);
 	}
-	close(sendfh);
-
-	error:
-	sqlite3_free_table(result);
+	close(fd);
 }
 
 void
 SendResp_thumbnail(struct upnphttp * h, char * object)
 {
-	char header[1500];
-	char sql_buf[256];
-	char **result;
-	int rows = 0;
+	char header[512];
 	char *path;
 	char date[30];
 	time_t curtime = time(NULL);
+	int ret;
 	ExifData *ed;
 	ExifLoader *l;
 
-	memset(header, 0, 1500);
-
 	if( h->reqflags & FLAG_XFERSTREAMING || h->reqflags & FLAG_RANGE )
 	{
 		DPRINTF(E_WARN, L_HTTP, "Client tried to specify transferMode as Streaming with an image!\n");
@@ -1306,88 +1404,86 @@ SendResp_thumbnail(struct upnphttp * h, char * object)
 	}
 
 	strip_ext(object);
-	sprintf(sql_buf, "SELECT PATH from DETAILS where ID = '%s'", object);
-	sql_get_table(db, sql_buf, &result, &rows, NULL);
-	if( !rows )
+	path = sql_get_text_field(db, "SELECT PATH from DETAILS where ID = '%s'", object);
+	if( !path )
 	{
-		DPRINTF(E_WARN, L_HTTP, "%s not found, responding ERROR 404\n", object);
+		DPRINTF(E_WARN, L_HTTP, "DETAIL ID %s not found, responding ERROR 404\n", object);
 		Send404(h);
-		goto error;
+		return;
 	}
-	path = result[1];
 	DPRINTF(E_INFO, L_HTTP, "Serving thumbnail for ObjectId: %s [%s]\n", object, path);
 
-	if( access(path, F_OK) == 0 )
+	if( access(path, F_OK) != 0 )
 	{
-		strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
-
-		l = exif_loader_new();
-		exif_loader_write_file(l, path);
-		ed = exif_loader_get_data(l);
-		exif_loader_unref(l);
-
-		if( !ed || !ed->size )
-		{
-			Send404(h);
-			if( ed )
-				exif_data_unref(ed);
-			goto error;
-		}
-		sprintf(header, "HTTP/1.1 200 OK\r\n"
-				"Content-Type: image/jpeg\r\n"
-				"Content-Length: %d\r\n"
-				"Connection: close\r\n"
-				"Date: %s\r\n"
-				"EXT:\r\n"
-				"realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
-			 	"contentFeatures.dlna.org: DLNA.ORG_PN=JPEG_TN\r\n"
-				"Server: " MINIDLNA_SERVER_STRING "\r\n",
-				ed->size, date);
-
-		if( h->reqflags & FLAG_XFERBACKGROUND )
-		{
-			strcat(header, "transferMode.dlna.org: Background\r\n\r\n");
-		}
-		else //if( h->reqflags & FLAG_XFERINTERACTIVE )
-		{
-			strcat(header, "transferMode.dlna.org: Interactive\r\n\r\n");
-		}
-
-		if( (send_data(h, header, strlen(header), MSG_MORE) == 0) && (h->req_command != EHead) )
-		{
-			send_data(h, (char *)ed->data, ed->size, 0);
-		}
-		exif_data_unref(ed);
+		DPRINTF(E_ERROR, L_HTTP, "Error accessing %s\n", path);
+		sqlite3_free(path);
+		return;
 	}
+	strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
+
+	l = exif_loader_new();
+	exif_loader_write_file(l, path);
+	ed = exif_loader_get_data(l);
+	exif_loader_unref(l);
+	sqlite3_free(path);
+
+	if( !ed || !ed->size )
+	{
+		Send404(h);
+		if( ed )
+			exif_data_unref(ed);
+		return;
+	}
+	ret = snprintf(header, sizeof(header), "HTTP/1.1 200 OK\r\n"
+	                                       "Content-Type: image/jpeg\r\n"
+	                                       "Content-Length: %d\r\n"
+	                                       "Connection: close\r\n"
+	                                       "Date: %s\r\n"
+	                                       "EXT:\r\n"
+	                                       "realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
+	                                       "contentFeatures.dlna.org: DLNA.ORG_PN=JPEG_TN\r\n"
+	                                       "Server: " MINIDLNA_SERVER_STRING "\r\n"
+	                                       "transferMode.dlna.org: %s\r\n\r\n",
+	                                       ed->size, date,
+	                                       (h->reqflags & FLAG_XFERBACKGROUND) ? "Background" : "Interactive");
+
+	if( send_data(h, header, ret, MSG_MORE) == 0 )
+	{
+ 		if( h->req_command != EHead )
+			send_data(h, (char *)ed->data, ed->size, 0);
+	}
+	exif_data_unref(ed);
 	CloseSocket_upnphttp(h);
-	error:
-	sqlite3_free_table(result);
 }
 
 void
 SendResp_resizedimg(struct upnphttp * h, char * object)
 {
-	char header[1500];
+	char header[512];
 	char str_buf[256];
+	struct string_s str;
 	char **result;
 	char date[30];
 	char dlna_pn[4];
 	time_t curtime = time(NULL);
-	int width=640, height=480, dstw, dsth, rotation, size;
+	int width=640, height=480, dstw, dsth, size;
 	long srcw, srch;
 	unsigned char * data = NULL;
 	char *path, *file_path;
-	char *resolution, *tn;
+	char *resolution;
 	char *key, *val;
 	char *saveptr=NULL, *item=NULL;
+	/* Not implemented yet *
 	char *pixelshape=NULL;
+	int rotation; */
 	sqlite_int64 id;
-	int rows=0, chunked=0, ret;
+	int rows=0, chunked, ret;
 #ifdef __sparc__
+	char *tn;
 	ExifData *ed;
 	ExifLoader *l;
 #endif
-	image *imsrc = NULL, *imdst = NULL;
+	image_s *imsrc = NULL, *imdst = NULL;
 	int scale = 1;
 
 	id = strtoll(object, NULL, 10);
@@ -1414,7 +1510,6 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 #endif
 	file_path = result[3];
 	resolution = result[4];
-	tn = result[5];
 	srcw = strtol(resolution, &saveptr, 10);
 	srch = strtol(saveptr+1, NULL, 10);
 
@@ -1425,9 +1520,9 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 	}
 	while( item != NULL )
 	{
-		#ifdef TIVO_SUPPORT
+#ifdef TIVO_SUPPORT
 		decodeString(item, 1);
-		#endif
+#endif
 		val = item;
 		key = strsep(&val, "=");
 		DPRINTF(E_DEBUG, L_GENERAL, "%s: %s\n", key, val);
@@ -1439,6 +1534,7 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 		{
 			height = atoi(val);
 		}
+		/* Not implemented yet *
 		else if( strcasecmp(key, "rotation") == 0 )
 		{
 			rotation = atoi(val);
@@ -1446,7 +1542,7 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 		else if( strcasecmp(key, "pixelshape") == 0 )
 		{
 			pixelshape = val;
-		}
+		} */
 		item = strtok_r(NULL, "&,", &saveptr);
 	}
 	free(path);
@@ -1484,26 +1580,32 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 		scale = 2;
 
 	strftime(date, 30,"%a, %d %b %Y %H:%M:%S GMT" , gmtime(&curtime));
-	snprintf(header, sizeof(header)-100, "HTTP/1.1 200 OK\r\n"
-	                                     "Content-Type: image/jpeg\r\n"
-	                                     "Connection: close\r\n"
-	                                     "Date: %s\r\n"
-	                                     "EXT:\r\n"
-	                                     "realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
-	                                     "contentFeatures.dlna.org: DLNA.ORG_PN=JPEG_%s;DLNA.ORG_CI=1\r\n"
-	                                     "Server: " MINIDLNA_SERVER_STRING "\r\n",
-	                                     date, dlna_pn);
+
+        str.data = header;
+        str.size = sizeof(header);
+        str.off = 0;
+
+	strcatf(&str, "HTTP/1.1 200 OK\r\n"
+	              "Content-Type: image/jpeg\r\n"
+	              "Connection: close\r\n"
+	              "Date: %s\r\n"
+	              "EXT:\r\n"
+	              "realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
+	              "contentFeatures.dlna.org: DLNA.ORG_PN=JPEG_%s;DLNA.ORG_CI=1\r\n"
+	              "Server: " MINIDLNA_SERVER_STRING "\r\n",
+	              date, dlna_pn);
 	if( h->reqflags & FLAG_XFERINTERACTIVE )
 	{
-		strcat(header, "transferMode.dlna.org: Interactive\r\n");
+		strcatf(&str, "transferMode.dlna.org: Interactive\r\n");
 	}
 	else if( h->reqflags & FLAG_XFERBACKGROUND )
 	{
-		strcat(header, "transferMode.dlna.org: Background\r\n");
+		strcatf(&str, "transferMode.dlna.org: Background\r\n");
 	}
 
 	/* Resizing from a thumbnail is much faster than from a large image */
 #ifdef __sparc__
+	tn = result[5];
 	if( dstw <= 160 && dsth <= 120 && atoi(tn) )
 	{
 		l = exif_loader_new();
@@ -1526,12 +1628,13 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 #endif
 	if( strcmp(h->HttpVer, "HTTP/1.0") == 0 )
 	{
+		chunked = 0;
 		imsrc = image_new_from_jpeg(file_path, 1, NULL, 0, scale);
 	}
 	else
 	{
 		chunked = 1;
-		strcat(header, "Transfer-Encoding: chunked\r\n\r\n");
+		strcatf(&str, "Transfer-Encoding: chunked\r\n\r\n");
 	}
 
 	if( !chunked )
@@ -1546,11 +1649,10 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 		imdst = image_resize(imsrc, dstw, dsth);
 		data = image_save_to_jpeg_buf(imdst, &size);
 
-		sprintf(str_buf, "Content-Length: %d\r\n\r\n", size);
-		strcat(header, str_buf);
+		strcatf(&str, "Content-Length: %d\r\n\r\n", size);
 	}
 
-	if( (send_data(h, header, strlen(header), 0) == 0) && (h->req_command != EHead) )
+	if( (send_data(h, str.data, str.off, 0) == 0) && (h->req_command != EHead) )
 	{
 		if( chunked )
 		{
@@ -1590,8 +1692,8 @@ SendResp_resizedimg(struct upnphttp * h, char * object)
 void
 SendResp_dlnafile(struct upnphttp * h, char * object)
 {
-	char header[1500];
-	char hdr_buf[512];
+	char header[1024];
+	struct string_s str;
 	char sql_buf[256];
 	char **result;
 	int rows, ret;
@@ -1600,7 +1702,7 @@ SendResp_dlnafile(struct upnphttp * h, char * object)
 	off_t total, offset, size;
 	sqlite_int64 id;
 	int sendfh;
-	static struct { sqlite_int64 id; char path[PATH_MAX]; char mime[32]; char dlna[64]; } last_file = { 0 };
+	static struct { sqlite_int64 id; char path[PATH_MAX]; char mime[32]; char dlna[96]; } last_file = { 0 };
 #if USE_FORK
 	pid_t newpid = 0;
 #endif
@@ -1630,10 +1732,23 @@ SendResp_dlnafile(struct upnphttp * h, char * object)
 		{
 			strncpy(last_file.mime, result[4], sizeof(last_file.mime)-1);
 			/* From what I read, Samsung TV's expect a [wrong] MIME type of x-mkv. */
-			if( h->req_client == ESamsungTV )
+			if( h->reqflags & FLAG_SAMSUNG )
 			{
 				if( strcmp(last_file.mime+6, "x-matroska") == 0 )
 					strcpy(last_file.mime+8, "mkv");
+				/* Samsung TV's such as the A750 can natively support many
+				   Xvid/DivX AVI's however, the DLNA server needs the 
+				   mime type to say video/mpeg */
+				else if( h->req_client == ESamsungSeriesA &&
+				         strcmp(last_file.mime+6, "x-msvideo") == 0 )
+					strcpy(last_file.mime+6, "mpeg");
+			}
+			/* ... and Sony BDP-S370 won't play MKV unless we pretend it's a DiVX file */
+			else if( h->req_client == ESonyBDP )
+			{
+				if( strcmp(last_file.mime+6, "x-matroska") == 0 ||
+				    strcmp(last_file.mime+6, "mpeg") == 0 )
+					strcpy(last_file.mime+6, "divx");
 			}
 		}
 		else
@@ -1678,7 +1793,7 @@ SendResp_dlnafile(struct upnphttp * h, char * object)
 			DPRINTF(E_WARN, L_HTTP, "Client tried to specify transferMode as Interactive without an image!\n");
 			/* Samsung TVs (well, at least the A950) do this for some reason,
 			 * and I don't see them fixing this bug any time soon. */
-			if( h->req_client != ESamsungTV || GETFLAG(DLNA_STRICT_MASK) )
+			if( !(h->reqflags & FLAG_SAMSUNG) || GETFLAG(DLNA_STRICT_MASK) )
 			{
 				Send406(h);
 				goto error;
@@ -1691,17 +1806,26 @@ SendResp_dlnafile(struct upnphttp * h, char * object)
 	sendfh = open(last_file.path, O_RDONLY);
 	if( sendfh < 0 ) {
 		DPRINTF(E_ERROR, L_HTTP, "Error opening %s\n", last_file.path);
+		Send404(h);
 		goto error;
 	}
 	size = lseek(sendfh, 0, SEEK_END);
 	lseek(sendfh, 0, SEEK_SET);
 
-	sprintf(header, "HTTP/1.1 20%c OK\r\n"
-			"Content-Type: %s\r\n", (h->reqflags & FLAG_RANGE ? '6' : '0'), last_file.mime);
+	str.data = header;
+	str.size = sizeof(header);
+	str.off = 0;
+
+	strcatf(&str, "HTTP/1.1 20%c OK\r\n"
+	              "Content-Type: %s\r\n",
+	              (h->reqflags & FLAG_RANGE ? '6' : '0'),
+	              last_file.mime);
 	if( h->reqflags & FLAG_RANGE )
 	{
-		if( !h->req_RangeEnd )
-			h->req_RangeEnd = size;
+		if( !h->req_RangeEnd || h->req_RangeEnd == size )
+		{
+			h->req_RangeEnd = size - 1;
+		}
 		if( (h->req_RangeStart > h->req_RangeEnd) || (h->req_RangeStart < 0) )
 		{
 			DPRINTF(E_WARN, L_HTTP, "Specified range was invalid!\n");
@@ -1709,7 +1833,7 @@ SendResp_dlnafile(struct upnphttp * h, char * object)
 			close(sendfh);
 			goto error;
 		}
-		if( h->req_RangeEnd > size )
+		if( h->req_RangeEnd >= size )
 		{
 			DPRINTF(E_WARN, L_HTTP, "Specified range was outside file boundaries!\n");
 			Send416(h);
@@ -1717,81 +1841,63 @@ SendResp_dlnafile(struct upnphttp * h, char * object)
 			goto error;
 		}
 
-		if( h->req_RangeEnd < size )
-		{
-			total = h->req_RangeEnd - h->req_RangeStart + 1;
-			sprintf(hdr_buf, "Content-Length: %jd\r\n"
-					 "Content-Range: bytes %jd-%jd/%jd\r\n",
-					 total, h->req_RangeStart, h->req_RangeEnd, size);
-		}
-		else
-		{
-			h->req_RangeEnd = size;
-			total = size - h->req_RangeStart;
-			sprintf(hdr_buf, "Content-Length: %jd\r\n"
-					 "Content-Range: bytes %jd-%jd/%jd\r\n",
-					 total, h->req_RangeStart, size-1, size);
-		}
+		total = h->req_RangeEnd - h->req_RangeStart + 1;
+		strcatf(&str, "Content-Length: %jd\r\n"
+		              "Content-Range: bytes %jd-%jd/%jd\r\n",
+		              (intmax_t)total, (intmax_t)h->req_RangeStart,
+		              (intmax_t)h->req_RangeEnd, (intmax_t)size);
 	}
 	else
 	{
-		h->req_RangeEnd = size;
+		h->req_RangeEnd = size - 1;
 		total = size;
-		sprintf(hdr_buf, "Content-Length: %jd\r\n", total);
+		strcatf(&str, "Content-Length: %jd\r\n", (intmax_t)total);
 	}
-	strcat(header, hdr_buf);
 
 	if( h->reqflags & FLAG_XFERSTREAMING )
 	{
-		strcat(header, "transferMode.dlna.org: Streaming\r\n");
+		strcatf(&str, "transferMode.dlna.org: Streaming\r\n");
 	}
 	else if( h->reqflags & FLAG_XFERBACKGROUND )
 	{
 		if( strncmp(last_file.mime, "image", 5) == 0 )
-			strcat(header, "transferMode.dlna.org: Background\r\n");
+			strcatf(&str, "transferMode.dlna.org: Background\r\n");
 	}
 	else //if( h->reqflags & FLAG_XFERINTERACTIVE )
 	{
 		if( (strncmp(last_file.mime, "video", 5) == 0) ||
 		    (strncmp(last_file.mime, "audio", 5) == 0) )
 		{
-			strcat(header, "transferMode.dlna.org: Streaming\r\n");
+			strcatf(&str, "transferMode.dlna.org: Streaming\r\n");
 		}
 		else
 		{
-			strcat(header, "transferMode.dlna.org: Interactive\r\n");
+			strcatf(&str, "transferMode.dlna.org: Interactive\r\n");
 		}
 	}
 
 	if( h->reqflags & FLAG_CAPTION )
 	{
-		sprintf(sql_buf, "SELECT 1 from CAPTIONS where ID = '%lld'", id);
-		ret = sql_get_table(db, sql_buf, &result, &rows, NULL);
-		if( ret == SQLITE_OK )
+		if( sql_get_int_field(db, "SELECT ID from CAPTIONS where ID = '%lld'", id) > 0 )
 		{
-			if( rows )
-			{
-				sprintf(hdr_buf, "CaptionInfo.sec: http://%s:%d/Captions/%lld.srt\r\n",
-				                 lan_addr[0].str, runtime_vars.port, id);
-				strcat(header, hdr_buf);
-			}
-			sqlite3_free_table(result);
+			strcatf(&str, "CaptionInfo.sec: http://%s:%d/Captions/%lld.srt\r\n",
+			              lan_addr[h->iface].str, runtime_vars.port, id);
 		}
 	}
 
-	sprintf(hdr_buf, "Accept-Ranges: bytes\r\n"
-			 "Connection: close\r\n"
-			 "Date: %s\r\n"
-			 "EXT:\r\n"
-	                 "realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
-			 "contentFeatures.dlna.org: %s\r\n"
-			 "Server: " MINIDLNA_SERVER_STRING "\r\n\r\n",
-			 date, last_file.dlna);
-	strcat(header, hdr_buf);
+	strcatf(&str, "Accept-Ranges: bytes\r\n"
+	              "Connection: close\r\n"
+	              "Date: %s\r\n"
+	              "EXT:\r\n"
+	              "realTimeInfo.dlna.org: DLNA.ORG_TLAG=*\r\n"
+	              "contentFeatures.dlna.org: %s\r\n"
+	              "Server: " MINIDLNA_SERVER_STRING "\r\n\r\n",
+	              date, last_file.dlna);
 
-	if( (send_data(h, header, strlen(header), MSG_MORE) == 0) && (h->req_command != EHead) && (sendfh > 0) )
+	if( send_data(h, str.data, str.off, MSG_MORE) == 0 )
 	{
-		send_file(h, sendfh, offset, h->req_RangeEnd);
+ 		if( h->req_command != EHead )
+			send_file(h, sendfh, offset, h->req_RangeEnd);
 	}
 	close(sendfh);
 
